@@ -1,7 +1,7 @@
 /*************************************************************************/
 /* Copyright (c) 2004                                                    */
 /* Daniel Sleator, David Temperley, and John Lafferty                    */
-/* Copyright 2018, 2019, Amir Plivatsky                                  */
+/* Copyright 2018-2020, Amir Plivatsky                                  */
 /* All rights reserved                                                   */
 /*                                                                       */
 /* Use of the link grammar parsing system is subject to the terms of the */
@@ -11,17 +11,20 @@
 /*                                                                       */
 /*************************************************************************/
 #include <string.h>
-#include <limits.h>                     // UINT_MAX
 
 #include "api-structures.h"             // Sentence
 #include "connectors.h"
-#include "tracon-set.h"
+#include "dict-common/dict-structures.h"
+#include "dict-common/dict-utils.h"     // copy_Exp
+#include "dict-common/regex-morph.h"    // match_regex
 #include "disjunct-utils.h"
-#include "print/print-util.h"
 #include "memory-pool.h"
-#include "utilities.h"
+#include "prepare/build-disjuncts.h"
+#include "print/print-util.h"
 #include "tokenize/tok-structures.h"    // XXX TODO provide gword access methods!
 #include "tokenize/word-structures.h"
+#include "tracon-set.h"
+#include "utilities.h"
 
 /* Disjunct utilities ... */
 
@@ -84,24 +87,21 @@ unsigned int count_disjuncts(Disjunct * d)
 	return count;
 }
 
-/** Returns the number of disjuncts and connectors in the sentence. */
-static void count_disjuncts_and_connectors(Sentence sent,
-                                           unsigned int *dca, unsigned int *cca)
+/** Returns the number of connectors in the sentence. */
+static unsigned int count_connectors(Sentence sent)
 {
-	unsigned int ccnt = 0, dcnt = 0;
+	unsigned int ccnt = 0;
 
 	for (WordIdx w = 0; w < sent->length; w++)
 	{
 		for (Disjunct *d = sent->word[w].d; d != NULL; d = d->next)
 		{
-			dcnt++;
 			for (Connector *c = d->left; c != NULL; c = c->next) ccnt++;
 			for (Connector *c = d->right; c !=NULL; c = c->next) ccnt++;
 		}
 	}
 
-	*cca = ccnt;
-	*dca = dcnt;
+	return ccnt;
 }
 /* ============================================================= */
 
@@ -122,11 +122,11 @@ static inline unsigned int old_hash_disjunct(disjunct_dup_table *dt, Disjunct *
 {
 	unsigned int i;
 	i = 0;
-	for (Connector *e = d->left ; e != NULL; e = e->next) {
-		i = (5 * (i + e->desc->uc_num)) + (unsigned int)e->desc->lc_letters + 7;
+	for (Connector *e = d->left; e != NULL; e = e->next) {
+		i = (41 * (i + e->desc->uc_num)) + (unsigned int)e->desc->lc_letters + 7;
 	}
-	for (Connector *e = d->right ; e != NULL; e = e->next) {
-		i = (5 * (i + e->desc->uc_num)) + (unsigned int)e->desc->lc_letters + 7;
+	for (Connector *e = d->right; e != NULL; e = e->next) {
+		i = (41 * (i + e->desc->uc_num)) + (unsigned int)e->desc->lc_letters + 7;
 	}
 #if 0 /* Redundant - the connector hashing has enough entropy. */
 	i += string_hash(d->word_string);
@@ -195,58 +195,6 @@ static bool disjuncts_equal(Disjunct * d1, Disjunct * d2)
 }
 #endif
 
-#if 0
-/**
- * Duplicate the given connector chain.
- * If the argument is NULL, return NULL.
- */
-static Connector *connectors_dup(Pool_desc *connector_pool, Connector *origc)
-{
-	Connector head;
-	Connector *prevc = &head;
-	Connector *newc = &head;
-
-	for (Connector *t = origc; t != NULL;  t = t->next)
-	{
-		newc = connector_new(connector_pool, NULL, NULL);
-		*newc = *t;
-
-		prevc->next = newc;
-		prevc = newc;
-	}
-	newc->next = NULL;
-
-	return head.next;
-}
-
-/**
- * Duplicate the given disjunct chain.
- * If the argument is NULL, return NULL.
- */
-static Disjunct *disjuncts_dup(Pool_desc *Disjunct_pool, Pool_desc *Connector_pool,
-                        Disjunct *origd)
-{
-	Disjunct head;
-	Disjunct *prevd = &head;
-	Disjunct *newd = &head;
-
-	for (Disjunct *t = origd; t != NULL; t = t->next)
-	{
-		newd = pool_alloc(Disjunct_pool);
-		newd->word_string = t->word_string;
-		newd->cost = t->cost;
-		newd->left = connectors_dup(Connector_pool, t->left);
-		newd->right = connectors_dup(Connector_pool, t->right);
-		newd->originating_gword = t->originating_gword;
-		prevd->next = newd;
-		prevd = newd;
-	}
-	newd->next = NULL;
-
-	return head.next;
-}
-#endif
-
 static disjunct_dup_table * disjunct_dup_table_new(size_t sz)
 {
 	disjunct_dup_table *dt;
@@ -350,13 +298,17 @@ Disjunct *eliminate_duplicate_disjuncts(Disjunct *dw)
 {
 	unsigned int count = 0;
 	disjunct_dup_table *dt;
+	/* This initialization is unneeded because the first disjunct is never
+	 * eliminated. However, omitting it generates "uninitialized" compiler
+	 * warning. Setting it to NULL generates clang-analyzer error on
+	 * possible NULL dereference. */
+	Disjunct *prev = dw;
 
 	dt = disjunct_dup_table_new(next_power_of_two_up(2 * count_disjuncts(dw)));
 
-	for (Disjunct **dd = &dw; *dd != NULL; /* See: NEXT */)
+	for (Disjunct *d = dw; d != NULL; d = d->next)
 	{
 		Disjunct *dx;
-		Disjunct *d = *dd;
 		unsigned int h = old_hash_disjunct(dt, d);
 
 		for (dx = dt->dup_table[h]; dx != NULL; dx = dx->dup_table_next)
@@ -374,13 +326,13 @@ Disjunct *eliminate_duplicate_disjuncts(Disjunct *dw)
 				gword_set_union(dx->originating_gword, d->originating_gword);
 
 			count++;
-			*dd = d->next; /* NEXT - set current disjunct to the next one. */
+			prev->next = d->next;
 		}
 		else
 		{
 			d->dup_table_next = dt->dup_table[h];
 			dt->dup_table[h] = d;
-			dd = &d->next; /* NEXT */
+			prev = d;
 		}
 	}
 
@@ -444,83 +396,25 @@ int right_connector_count(Disjunct * d)
 	return i;
 }
 
-/* ============================================================= */
-
-/**
- * Record the wordgraph word to which the X-node belongs, in each of its
- * disjuncts.
- */
-void word_record_in_disjunct(const Gword * gw, Disjunct * d)
+/** Returns the number of disjuncts and connectors in the sentence. */
+void count_disjuncts_and_connectors(Sentence sent, unsigned int *dca,
+                                    unsigned int *cca)
 {
-	for (;d != NULL; d=d->next) {
-		d->originating_gword = (gword_set *)&gw->gword_set_head;
-	}
-}
+	unsigned int ccnt = 0, dcnt = 0;
 
-
-/* ================ Pack disjuncts and connectors ============== */
-/* Print one connector with all the details.
- * mCnameD<tracon_id>(nearest_word, length_limit)x
- * optional m: "@" for multi (else nothing)
- * Cname: Connector name
- * Optional D: "-" / "+" (if dir != -1)
- * Optional <tracon_id>: tracon_id (if not 0)
- * Optional (nearest_word, length_limit): if both are not 0
- * x: Shallow/deep indication as "s" / "d" (if shallow != -1)
- */
-void print_one_connector(Connector * e, int dir, int shallow)
-{
-	printf("%s%s", e->multi ? "@" : "", connector_string(e));
-	if (-1 != dir) printf("%c", "-+"[dir]);
-	if (e->tracon_id)
+	for (WordIdx w = 0; w < sent->length; w++)
 	{
-		if (e->refcount)
-			printf("<%d,%d>", e->tracon_id, e->refcount);
-		else
-			printf("<%d>", e->tracon_id);
-	}
-	if ((0 != e->nearest_word) || (0 != e->length_limit))
-		printf("(%d,%d)", e->nearest_word, e->length_limit);
-	if (-1 != shallow)
-		printf("%c", (0 == shallow) ? 'd' : 's');
-}
-
-void print_connector_list(Connector * e)
-{
-	for (;e != NULL; e=e->next)
-	{
-		print_one_connector(e, /*dir*/-1, /*shallow*/-1);
-		if (e->next != NULL) printf(" ");
-	}
-}
-
-void print_disjunct_list(Disjunct * dj)
-{
-	int i = 0;
-	char word[MAX_WORD + 32];
-
-	for (;dj != NULL; dj=dj->next)
-	{
-		lg_strlcpy(word, dj->word_string, sizeof(word));
-		patch_subscript_mark(word);
-		printf("%16s: ", word);
-		printf("[%d](%4.2f) ", i++, dj->cost);
-		print_connector_list(dj->left);
-		printf(" <--> ");
-		print_connector_list(dj->right);
-		printf("\n");
-	}
-}
-
-void print_all_disjuncts(Sentence sent)
-{
-		for (WordIdx w = 0; w < sent->length; w++)
+		for (Disjunct *d = sent->word[w].d; d != NULL; d = d->next)
 		{
-			printf("Word %zu:\n", w);
-			print_disjunct_list(sent->word[w].d);
+			dcnt++;
+			for (Connector *c = d->left; c != NULL; c = c->next) ccnt++;
+			for (Connector *c = d->right; c !=NULL; c = c->next) ccnt++;
 		}
-}
+	}
 
+	*cca = ccnt;
+	*dca = dcnt;
+}
 
 /* ============= Connector encoding, sharing and packing ============= */
 
@@ -544,12 +438,9 @@ void print_all_disjuncts(Sentence sent)
  * comparison.
  *
  * For the parsing step, identical tracons are assigned a unique tracon
- * ID, which is kept in their first connector. The rest of their
- * connectors also have tracon IDs, which belong to tracons starting
- * with that connectors.
- *
- * For the pruning step, all the tracon IDs are set to 0 (see below for
- * how tracon_id is used during the power pruning).
+ * ID, which is kept in their first connector tracon_id field. The rest of
+ * their connectors also have tracon IDs, which belong to tracons starting
+ * with that connectors. The tracon_id is not used for pruning.
  *
  * For the pruning step, more things are done:
  * Additional data structure - a tracon list - is constructed, which
@@ -568,17 +459,17 @@ void print_all_disjuncts(Sentence sent)
  * The first connector of each tracon is inserted into the power table,
  * along with its reference count. When a connector cannot be matched,
  * this means that all the disjuncts that contain its tracon also cannot
- * be matched. When it is marked as bad (cannot match) or good (match
- * found), due to the tracon memory sharing all the connectors that
- * share the same memory are marked simultaneously, and thus are
- * detected when the next disjuncts are examined without a need to
- * further process them. This saves much processing and also drastically
- * reduces the "power_cost". Setting the nearest_word field is hence
- * done only once per tracon on each pass. The tracon IDs are used to
- * detect already-processed tracons - they are assigned the pass number
- * so each tracon is processed only once per pass. The connector
- * refcount field is used to discard connectors from the power table
- * when all the disjuncts that contain them are discarded.
+ * be matched. It is then marked as bad (by nearest_word=BAD_WORD) and
+ * due to the tracon memory sharing all the connectors that share the same
+ * memory are marked simultaneously, and thus are detected when the next
+ * disjuncts are examined without a need to further process them.
+ * This drastically reduces the "power_cost" and saves much processing.
+ * Setting the nearest_word field is hence done only once per tracon on
+ * each pass. The pass_number field is used to detect already-processed
+ * good tracons - they are assigned the pass number so each tracon is
+ * processed only once per pass. The connector refcount field is used to
+ * discard connectors from the power table when all the disjuncts that
+ * contain them are discarded.
  *
  * PP pruning:
  * Here too only the first connector in each tracon needs to be
@@ -683,7 +574,7 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 	{
 		newc = NULL;
 
-		if ((NULL != ts) && (NULL != ts->csid[dir]))
+		if (NULL != ts->csid[dir])
 		{
 			/* Encoding is used - share tracons. */
 			Connector **tracon = tracon_set_add(o, ts->csid[dir]);
@@ -693,7 +584,7 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 				/* The first time we encounter this tracon. */
 				*tracon = lcblock; /* Save its future location in the tracon_set. */
 
-				if (ts->is_pruning)
+				if (NULL != tl)
 				{
 					tlsz_check(tl, tl->entries[dir], dir);
 					uint32_t cblock_index = (uint32_t)(lcblock - ts->cblock_base);
@@ -706,7 +597,8 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 				newc = *tracon;
 				if (!ts->is_pruning)
 				{
-					if (o->nearest_word != newc->nearest_word)
+					if ((o->nearest_word != newc->nearest_word) ||
+					    (o->farthest_word != newc->farthest_word))
 					{
 						/* This is a rare case in which a shallow and deep
 						 * connectors don't have the same nearest_word, because
@@ -714,11 +606,12 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 						 * earlier. Because the nearest word is different, we
 						 * cannot share it. (Such shallow and deep tracons could
 						 * be shared separately, but because this is a rare
-						 * event there is no need to do that.)
+						 * event there is no benefit to do that.)
 						 * Note:
 						 * In case the parsing ever depends on other Connector
 						 * fields, their will be a need to add a check for them
-						 * here. */
+						 * here.
+						 * Update: farthest_word added. */
 						newc = NULL; /* Don't share it. */
 					}
 				}
@@ -733,9 +626,8 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 
 			if (ts->is_pruning)
 			{
-				/* Initializations for the pruning step. */
-				newc->refcount = 1;  /* The first connector at this location. */
-				newc->tracon_id = 0; /* Used in power_prune() for pass number. */
+				/* Initialize for the pruning step when no sharing is done yet. */
+				newc->refcount = 1;  /* No sharing yet. */
 				if (NULL != tl)
 					tl->num_cnctrs_per_word[dir][w]++;
 			}
@@ -743,14 +635,11 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 			{
 				/* For the parsing step we need a unique ID. */
 				newc->tracon_id = ts->next_id[dir]++;
-#ifdef DEBUG
-				newc->refcount = 0;  /* Not used; zero for debug consistency. */
-#endif
 			}
 		}
 		else
 		{
-			if (ts->is_pruning)
+			if (NULL != tl)
 			{
 				for (Connector *n = newc; NULL != n; n = n->next)
 					n->refcount++;
@@ -771,72 +660,49 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 	return head.next;
 }
 
-/**
- * Set dummy tracon_id's.
- * To be used for short sentences (for which a full encoding is too
- * costly) or for library tests that actually bypass the use of tracon IDs
- * (to validate that the tracon_id implementation didn't introduce bugs).
- */
-static int enumerate_connectors_sequentially(Tracon_sharing *ts)
-{
-	int id = ts->word_offset;
-
-	for (size_t i = 0; i < ts->num_disjuncts; i++)
-	{
-		Disjunct *d = &ts->dblock_base[i];
-
-		for (Connector *c = d->left; NULL != c; c = c->next)
-			c->tracon_id = id++;
-
-		for (Connector *c = d->right; NULL != c; c = c->next)
-			c->tracon_id = id++;
-
-	}
-
-	return id + 1;
-}
-
 /**
  * Pack the given disjunct chain in a contiguous memory block.
  * If the disjunct is NULL, return NULL.
  */
-static Disjunct *pack_disjuncts(Tracon_sharing *ts, Disjunct *origd, int w)
+static Disjunct *pack_disjunct(Tracon_sharing *ts, Disjunct *d, int w)
+{
+	Disjunct *newd;
+	uintptr_t token = (uintptr_t)w;
+
+	newd = (ts->dblock)++;
+	newd->word_string = d->word_string;
+	newd->cost = d->cost;
+	newd->originating_gword = d->originating_gword;
+
+	if (NULL == ts->tracon_list)
+		 token = (uintptr_t)d->originating_gword;
+
+	if ((token != ts->last_token) && (NULL != ts->csid[0]))
+	{
+		ts->last_token = token;
+		//printf("Token %ld\n", token);
+		tracon_set_reset(ts->csid[0]);
+		tracon_set_reset(ts->csid[1]);
+	}
+	newd->left = pack_connectors(ts, d->left, 0, w);
+	newd->right = pack_connectors(ts, d->right, 1,  w);
+
+	return newd;
+}
+
+static Disjunct *pack_disjuncts(Sentence sent, Tracon_sharing *ts,
+                                Disjunct *origd, int w)
 {
 	Disjunct head;
 	Disjunct *prevd = &head;
-	Disjunct *newd = &head;
-	Disjunct *ldblock = ts->dblock; /* For convenience. */
-	uintptr_t token = (uintptr_t)w;
 
-	for (Disjunct *t = origd; NULL != t; t = t->next)
+	for (Disjunct *d = origd; NULL != d; d = d->next)
 	{
-		newd = ldblock++;
-		newd->word_string = t->word_string;
-		newd->cost = t->cost;
-		newd->originating_gword = t->originating_gword;
-
-		if (!ts->is_pruning)
-		    token = (uintptr_t)t->originating_gword;
-
-		if (token != ts->last_token)
-		{
-			ts->last_token = token;
-			if (NULL != ts->csid[0])
-			{
-				//printf("Token %ld\n", token);
-				tracon_set_reset(ts->csid[0]);
-				tracon_set_reset(ts->csid[1]);
-			}
-		}
-		newd->left = pack_connectors(ts, t->left, 0, w);
-		newd->right = pack_connectors(ts, t->right, 1,  w);
-
-		prevd->next = newd;
-		prevd = newd;
+		prevd->next = pack_disjunct(ts, d, w);
+		prevd = prevd->next;
 	}
-	newd->next = NULL;
+	prevd->next = NULL;
 
-	ts->dblock = ldblock;
 	return head.next;
 }
 
@@ -868,28 +734,38 @@ static Disjunct *pack_disjuncts(Tracon_sharing *ts, Disjunct *origd, int w)
  *   starting with a shallow connector will be considered different than
  *   similar ones starting with a deep connector.
  *
+ * Note:
+ * In order to save overhead, sentences shorter than
+ * sent->min_len_encoding don't undergo encoding - only packing.
+ * This can also be used for library tests that totally bypass the use of
+ * connector encoding (to validate that the tracon_id/sharing/refcount
+ * implementation didn't introduce bugs in the pruning and parsing steps).
+ * E.g. when using link-parser:
+ * - To entirely disable connector encoding:
+ * link-parser -test=min-len-encoding:254
+ * - To use connector encoding even for short sentences:
+ * link-parser -test=min-len-encoding:0
+ * Any different result (e.g. number of discarded disjuncts in the pruning
+ * step or different parsing results) indicates a bug.
+ *
  * @param is_pruning TRUE if invoked for pruning, FALSE if invoked for parsing.
  * @return The said context descriptor.
  */
-static Tracon_sharing *pack_sentence_init(Sentence sent, bool is_pruning,
-                                          bool do_encoding)
+static Tracon_sharing *pack_sentence_init(Sentence sent, bool is_pruning)
 {
-	unsigned int dcnt = 0;
-	unsigned int ccnt = 0;
-	Tracon_sharing *ts;
-
+	unsigned int dcnt = 0, ccnt = 0;
 	count_disjuncts_and_connectors(sent, &dcnt, &ccnt);
 
-#define CONN_ALIGNMENT sizeof(Connector)
 	size_t dsize = dcnt * sizeof(Disjunct);
-	dsize = ALIGN(dsize, CONN_ALIGNMENT); /* Align connector block. */
+	if (sizeof(Disjunct) != 64)
+		dsize = ALIGN(dsize, sizeof(Connector));
 	size_t csize = ccnt * sizeof(Connector);
 	size_t memblock_sz = dsize + csize;
 	void *memblock = malloc(memblock_sz);
 	Disjunct *dblock = memblock;
 	Connector *cblock = (Connector *)((char *)memblock + dsize);
 
-	ts = malloc(sizeof(Tracon_sharing));
+	Tracon_sharing *ts = malloc(sizeof(Tracon_sharing));
 	memset(ts, 0, sizeof(Tracon_sharing));
 
 	ts->memblock = memblock;
@@ -900,11 +776,12 @@ static Tracon_sharing *pack_sentence_init(Sentence sent, bool is_pruning,
 	ts->num_connectors = ccnt;
 	ts->num_disjuncts = dcnt;
 	ts->word_offset = is_pruning ? 1 : WORD_OFFSET;
+	ts->is_pruning = is_pruning;
 	ts->next_id[0] = ts->next_id[1] = ts->word_offset;
 	ts->last_token = (uintptr_t)-1;
-	ts->is_pruning = is_pruning;
 
-	if (do_encoding)
+	/* Encode connectors only for long-enough sentences. */
+	if (sent->length >= sent->min_len_encoding)
 	{
 		ts->csid[0] = tracon_set_create();
 		ts->csid[1] = tracon_set_create();
@@ -925,6 +802,15 @@ static Tracon_sharing *pack_sentence_init(Sentence sent, bool is_pruning,
 		}
 	}
 
+	if (!is_pruning && (ts->memblock != sent->dc_memblock))
+	{
+		/* The disjunct & connector content is stored in dc_memblock.
+		 * It will be freed at sentence_delete(). */
+		if (sent->dc_memblock) free(sent->dc_memblock);
+		sent->dc_memblock = ts->memblock;
+		sent->num_disjuncts = ts->num_disjuncts;
+	}
+
 	return ts;
 }
 
@@ -946,11 +832,10 @@ void free_tracon_sharing(Tracon_sharing *ts)
 		}
 	}
 
-	if (NULL != ts->tracon_list)
-	{
-		free(ts->tracon_list);
-		ts->tracon_list = NULL;
-	}
+	if (NULL != ts->d) free(ts->d);
+	free(ts->tracon_list);
+	ts->tracon_list = NULL;
+
 	free(ts);
 }
 
@@ -961,55 +846,35 @@ void free_tracon_sharing(Tracon_sharing *ts)
  *
  * The disjunct and connectors packing in a contiguous memory facilitate a
  * better memory caching for long sentences (a performance gain of a few
- * percents).
+ * percents in the initial implementation, in which this was the sole
+ * purpose of this packing.) In addition, tracon memory sharing
+ * drastically reduces the memory used for connectors.
  *
  * The tracon IDs (if invoked for the parsing step) or tracon lists (if
  * invoked for pruning step) allow for a huge performance boost at these
  * steps.
- *
- * Note:
- * In order to save overhead, sentences shorter than
- * sent->min_len_encoding don't undergo encoding - only packing.
- * This can also be used for library tests that totally bypass the use of
- * connector encoding (to validate that the
- * tracon_id/packing/sharing/refcount implementation didn't introduce bugs
- * in the pruning and parsing steps).
- * E.g. when using link-parser:
- * - To entirely disable connector encoding:
- * link-parser -test=len-trailing-hash:254
- * - To use connector encoding even for short sentences:
- * link-parser -test=len-trailing-hash:0
- * Any different result (e.g. number of discarded disjuncts in the pruning
- * step or different parsing results) indicates a bug.
  */
 static Tracon_sharing *pack_sentence(Sentence sent, bool is_pruning)
 {
-	bool do_encoding = sent->length >= sent->min_len_encoding;
-	Tracon_sharing *ts;
-
-	ts = pack_sentence_init(sent, is_pruning, do_encoding);
+	Tracon_sharing *ts = pack_sentence_init(sent, is_pruning);
 
 	for (WordIdx w = 0; w < sent->length; w++)
 	{
-		sent->word[w].d = pack_disjuncts(ts, sent->word[w].d, w);
+		sent->word[w].d = pack_disjuncts(sent, ts, sent->word[w].d, w);
 	}
 
-	if (is_pruning && !do_encoding)
-	{
-		lgdebug(D_DISJ, "enumerate_connectors_sequentially\n");
-		enumerate_connectors_sequentially(ts);
-	}
-
-	/* On long sentences, many MB of connector-space are saved, but we
-	 * cannot use a realloc() here without the overhead of relocating
-	 * the pointers in the used part of memblock (if realloc() returns a
-	 * different address). */
-
 	return ts;
 }
 
+/**
+ * Pack the sentence for pruning.
+ * @return New tracon sharing descriptor.
+ */
 Tracon_sharing *pack_sentence_for_pruning(Sentence sent)
 {
+	unsigned int ccnt_before = 0;
+	if (verbosity_level(D_DISJ)) ccnt_before = count_connectors(sent);
+
 	Tracon_sharing *ts = pack_sentence(sent, true);
 
 	if (NULL == ts->csid[0])
@@ -1020,20 +885,33 @@ Tracon_sharing *pack_sentence_for_pruning(Sentence sent)
 	else
 	{
 		lgdebug(D_DISJ, "Debug: Encode for pruning (len %zu): "
-		        "tracon_id %zu (%zu+,%zu-), shared connectors %ld\n",
+		        "tracon_id %zu (%zu+,%zu-), shared connectors %d\n",
 		        sent->length,
 		        ts->tracon_list->entries[0]+ts->tracon_list->entries[1],
 		        ts->tracon_list->entries[0], ts->tracon_list->entries[1],
-				  &ts->cblock_base[ts->num_connectors] - ts->cblock);
+				  (int)(&ts->cblock_base[ccnt_before] - ts->cblock));
 	}
 
 	return ts;
 }
 
+/**
+ * Pack the sentence for parsing.
+ * @return New tracon sharing descriptor.
+ */
 Tracon_sharing *pack_sentence_for_parsing(Sentence sent)
 {
+	unsigned int ccnt_before = 0;
+	if (verbosity_level(D_DISJ)) ccnt_before = count_connectors(sent);
+
 	Tracon_sharing *ts = pack_sentence(sent, false);
 
+	if (verbosity_level(D_SPEC+2))
+	{
+		printf("pack_sentence_for_parsing (null_count %u):\n", sent->null_count);
+		print_all_disjuncts(sent);
+	}
+
 	if (NULL == ts->csid[0])
 	{
 		lgdebug(D_DISJ, "Debug: Encode for parsing (len %zu): None\n",
@@ -1042,35 +920,36 @@ Tracon_sharing *pack_sentence_for_parsing(Sentence sent)
 	else
 	{
 		lgdebug(D_DISJ, "Debug: Encode for parsing (len %zu): "
-		        "tracon_id %d (%d+,%d-), shared connectors %ld\n",
+		        "tracon_id %d (%d+,%d-), shared connectors %d\n",
 		        sent->length,
 		        (ts->next_id[0]-ts->word_offset)+(ts->next_id[1]-ts->word_offset),
 		        ts->next_id[0]-ts->word_offset, ts->next_id[1]-ts->word_offset,
-		        &ts->cblock_base[ts->num_connectors] - ts->cblock);
+		        (int)(&ts->cblock_base[ccnt_before] - ts->cblock));
 	}
 
 	return ts;
 }
 
 /* ============ Save and restore sentence disjuncts ============ */
-void *save_disjuncts(Sentence sent, Tracon_sharing *ts, Disjunct **disjuncts)
+void *save_disjuncts(Sentence sent, Tracon_sharing *ts)
 {
-	for (WordIdx w = 0; w < sent->length; w++)
-		disjuncts[w] = sent->word[w].d;
-
 	void *saved_memblock = malloc(ts->memblock_sz);
 	memcpy(saved_memblock, ts->memblock, ts->memblock_sz);
 
+	if (NULL == ts->d)
+		ts->d = malloc(sent->length * sizeof(Disjunct *));
+	for (WordIdx w = 0; w < sent->length; w++)
+		ts->d[w] = sent->word[w].d;
+
 	return saved_memblock;
 }
 
-void restore_disjuncts(Sentence sent, Disjunct **disjuncts,
-                       void *saved_memblock, Tracon_sharing *ts)
+void restore_disjuncts(Sentence sent, void *saved_memblock, Tracon_sharing *ts)
 {
 	if (NULL == saved_memblock) return;
 
 	for (WordIdx w = 0; w < sent->length; w++)
-		sent->word[w].d = disjuncts[w];
+		sent->word[w].d = ts->d[w];
 
 	memcpy(ts->memblock, saved_memblock, ts->memblock_sz);
 }
